# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import concurrent.futures
import functools
import sys
import unittest

def isolation(f):
    if 'coverage' in sys.modules:
        # Process isolation would break coverage measurements.
        # Oh well. FIXME.
        return f
    else:
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
                ftr = executor.submit(f, *args, **kwargs)
                return ftr.result()
    return wrapper

def testcase(f):
    class TestCase(unittest.TestCase):  # pylint: disable=redefined-outer-name
        @staticmethod
        def test():
            return f()
        def __str__(self):
            return f'{f.__module__}.{f.__name__}'
    return TestCase

tc = unittest.TestCase('__hash__')

assert_equal = tc.assertEqual
assert_is = tc.assertIs
assert_is_instance = tc.assertIsInstance
assert_not_equal = tc.assertNotEqual

del tc

class TestCase(unittest.TestCase):
    def __str__(self):
        return '{cls}.{name}'.format(
            cls=unittest.util.strclass(self.__class__),
            name=self._testMethodName,
        )

__all__ = [
    'isolation',
    'testcase',
    'TestCase',
    # nose-compatible:
    'assert_equal',
    'assert_is',
    'assert_is_instance',
    'assert_not_equal',
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2021 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
I/O and encodings
'''

import codecs
import io
import sys

def enc_eq(e1, e2):
    '''
    check if two encodings are equal
    '''
    return (
        codecs.lookup(e1).name ==
        codecs.lookup(e2).name
    )

def get_encoding():
    '''
    get locale encoding (from sys.stdout);
    upgrade ASCII to UTF-8
    '''
    locale_encoding = sys.stdout.encoding
    if enc_eq(locale_encoding, 'ASCII'):
        return 'UTF-8'
    else:
        return locale_encoding

def open_file(path, *, encoding, errors):
    '''
    open() with special case for "-"
    '''
    if path == '-':
        return io.TextIOWrapper(
            sys.stdin.buffer,
            encoding=encoding,
            errors=errors,
        )
    else:
        return open(  # pylint: disable=consider-using-with
            path, 'rt',
            encoding=encoding,
            errors=errors,
        )

__all__ = [
    'get_encoding',
    'open_file',
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
interface to eSpeak (NG)
'''

import ctypes
import itertools

try:
    _shlib = ctypes.CDLL('libespeak-ng.so.1')
    ng = True
except OSError:  # no coverage
    _shlib = ctypes.CDLL('libespeak.so.1')
    ng = False

# const char *espeak_Info(const char **path_data)
_info = _shlib.espeak_Info
_info.argtypes = [ctypes.POINTER(ctypes.c_char_p)]
_info.restype = ctypes.c_char_p

def info():
    '''
    return eSpeak version information
    '''
    dummy = ctypes.c_char_p(b'')
    res = _info(ctypes.byref(dummy))
    return res.decode('ASCII')
version = info().split()[0]
del info

def vcmp(v1, v2):
    '''
    cmp()-style version comparison
    '''
    v1 = v1.split('.')
    v2 = v2.split('.')
    for c1, c2 in itertools.zip_longest(v1, v2, fillvalue=0):
        c1 = int(c1)
        c2 = int(c2)
        if c1 > c2:
            return 1
        elif c1 < c2:
            return -1
    return 0

# int espeak_Initialize(espeak_AUDIO_OUTPUT output, int buflength, const char *path, int options)
_initialize = _shlib.espeak_Initialize
_initialize.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_int]
_initialize.restype = ctypes.c_int

def init():
    '''
    initialize eSpeak
    '''
    rc = _initialize(0, 0, None, 0)
    if rc <= 0:
        raise RuntimeError('espeak_Initialize(): internal error')  # no coverage

# espeak_ERROR espeak_SetVoiceByName(const char *name)
_set_voice_by_name = _shlib.espeak_SetVoiceByName
_set_voice_by_name.argtypes = [ctypes.c_char_p]
_set_voice_by_name.restype = ctypes.c_int

def set_voice_by_name(s):
    '''
    use this voice for synthesis
    '''
    s = s.encode('ASCII')
    rc = _set_voice_by_name(s)
    if rc == 0:
        return
    else:  # no coverage
        if rc == -1:
            msg = 'internal error'
        elif rc == 1:
            msg = 'the command could not be buffered'
        else:
            msg = f'unknown error {rc}'
        raise RuntimeError('espeak_SetVoiceByName(): ' + msg)

if vcmp(version, '1.48.1') >= 0:

    # const char *espeak_TextToPhonemes(const void **textptr, int textmode, int phonememode)
    _text_to_phonemes = _shlib.espeak_TextToPhonemes
    _text_to_phonemes.restype = ctypes.c_char_p
    _text_to_phonemes.argtypes = [ctypes.POINTER(ctypes.c_char_p), ctypes.c_int, ctypes.c_int]

    def text_to_phonemes(s, *, ipa=False):
        '''
        translate text to phonemes
        '''
        s = s.encode('UTF-8')
        z = ctypes.c_char_p(s)
        zptr = ctypes.pointer(z)
        assert zptr.contents is not None
        if vcmp(version, '1.48.11') >= 0:
            ipa = ipa << 1
        else:
            ipa = ipa << 4  # no coverage
        res = _text_to_phonemes(zptr, 1, ipa)
        if zptr.contents.value is not None:
            raise RuntimeError  # no coverage
        return res.decode('UTF-8').strip()

elif vcmp(version, '1.47.08') >= 0:  # no coverage

    # void espeak_TextToPhonemes(const void *text, char *buffer, int size, int textmode, int phonememode)
    _text_to_phonemes = _shlib.espeak_TextToPhonemes
    _text_to_phonemes.restype = ctypes.c_char_p
    _text_to_phonemes.argtypes = [
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_char), ctypes.c_int,
        ctypes.c_int, ctypes.c_int
    ]

    def text_to_phonemes(s, *, ipa=False):
        '''
        translate text to phonemes
        '''
        s = s.encode('UTF-8')
        bufsize = 250
        buf = ctypes.create_string_buffer(bufsize)
        _text_to_phonemes(s, buf, bufsize, 1, ipa << 4)
        return buf.value.decode('UTF-8').strip()

else:  # no coverage

    raise RuntimeError('eSpeak >= 1.47.08 is required')

__all__ = [
    'init',
    'ng',
    'set_voice_by_name',
    'text_to_phonemes',
    'version',
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
anorack CLI
'''

import argparse
import io
import signal
import sys

from lib.articles import choose_art
from lib.io import (
    get_encoding,
    open_file,
)
from lib.misc import coerce_case, warn
from lib.parser import parse_file
from lib.phonetics import init as init_phonetics, text_to_phonemes
from lib.version import __version__

class VersionAction(argparse.Action):
    '''
    argparse --version action
    '''

    def __init__(self, option_strings, dest=argparse.SUPPRESS):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            help='show version information and exit'
        )

    def __call__(self, parser, namespace, values, option_string=None):
        from lib import espeak  # pylint: disable=import-outside-toplevel
        print(f'{parser.prog} {__version__}')
        print('+ Python {0}.{1}.{2}'.format(*sys.version_info))
        ng = ' NG' if espeak.ng else ''
        print(f'+ eSpeak{ng} {espeak.version}')
        parser.exit()

def check_word(loc, art, word, *, ipa=False):
    '''
    check if the word has correct article
    '''
    phon = text_to_phonemes(word, ipa=ipa)
    correct_art = choose_art(phon)
    if correct_art is NotImplemented:
        warn(f"can't determine correct article for {word!r} /{phon}/")
    elif art.lower() != correct_art:
        correct_art = coerce_case(art, correct_art)
        print(f'{loc}: {art} {word} -> {correct_art} {word} /{phon}/')

def main():
    '''
    run the program
    '''
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    ap = argparse.ArgumentParser(description='"a" vs "an" checker')
    ap.add_argument('--version', action=VersionAction)
    ap.add_argument('--ipa', action='store_true', help='use IPA instead of phoneme mnemonics')
    ap.add_argument('--traceback', action='store_true', help=argparse.SUPPRESS)
    ap.add_argument('files', metavar='FILE', nargs='*', default=['-'],
        help='file to check (default: stdin)')
    options = ap.parse_args()
    encoding = get_encoding()
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding, line_buffering=sys.stdout.line_buffering)
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding, line_buffering=sys.stdout.line_buffering)
    init_phonetics()
    rc = 0
    for path in options.files:
        try:
            file = open_file(path, encoding=encoding, errors='replace')
        except OSError as exc:
            if options.traceback:
                raise
            msg = f'{ap.prog}: {path}: {exc.strerror}'
            print(msg, file=sys.stderr)
            rc = 1
            continue
        with file:
            for loc, art, word in parse_file(file):
                check_word(loc, art, word, ipa=options.ipa)
    sys.exit(rc)

__all__ = ['main']

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
miscellanea
'''

import os
import sys

def warn(msg):
    '''
    print warning message
    '''
    prog = os.path.basename(sys.argv[0])
    print(f'{prog}: warning: {msg}', file=sys.stderr)

def _coerce_case(src, word):
    '''
    coerce word to the same case as src
    (simple version doesn't support title-case)
    '''
    if src.isupper():
        return word.upper()
    else:
        return word.lower()

def coerce_case(src, word):
    '''
    coerce word to the same case as src
    '''
    return (
        _coerce_case(src[:1], word[:1]) +
        _coerce_case(src[1:], word[1:])
    )

__all__ = [
    'warn',
    'coerce_case',
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
English phonetics
'''

import functools
import os

consonants = frozenset(
    'DNSTZbdfghjklmnprstvwz'
    'ðŋʃθʒbdfɡhjkɬmnpɹstvwz'
)
vowels = frozenset(
    '03@AEIOUVaeiou'
    'ɒɜəɑɛɪɔʊʌɐeiɔu'
)
accents = frozenset(
    ",'"
    "ˌˈ"
)

espeak = None
overrides = {}

def init():
    '''
    initialize underlying speech engine
    '''
    global espeak  # pylint: disable=global-statement,global-variable-not-assigned
    from lib import espeak  # pylint: disable=redefined-outer-name,import-outside-toplevel
    espeak.init()
    espeak.set_voice_by_name('en')
    here = os.path.dirname(__file__)
    # Ideally false positives should be fixed in eSpeak,
    # but as a stop-gap measure, we carry data file to correct some of them.
    path = f'{here}/../data/overrides'
    with open(path, 'rt', encoding='UTF-8') as file:
        for line in file:
            line = line.strip()
            (word, phon) = line.split('\t', 1)
            word = word.lower()
            overrides[word] = phon

@functools.lru_cache(maxsize=9999)
def text_to_phonemes(s, *, ipa=False):
    '''
    translate text to phonemes
    '''
    s = overrides.get(s.lower(), s)
    if s.startswith('[[') and s.endswith(']]'):
        return s.split('\t')[ipa][2:-2]
    else:
        return espeak.text_to_phonemes(s, ipa=ipa)

__all__ = [
    'consonants',
    'vowels',
    'accents',
    'text_to_phonemes',
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
English parser
'''

import re

class Location():
    '''
    location in a file
    '''
    def __init__(self, file, lineno):
        self.file = file
        self.lineno = lineno

    def __str__(self):
        return f'{self.file.name}:{self.lineno}'

find_articles = re.compile(
    r'''(?<!\w['’])(?:\b(an?)\s+(?:['‘"“]\s*)?(\d+|[^\W\d_]+)|\b(an?)\s*$)''',
    re.IGNORECASE
).finditer

def parse_file(file):
    '''
    parse the file:
    return sequence of (<location>, <article>, <word>) tuples
    '''
    carry = ''
    for i, line in enumerate(file, start=1):
        cline = carry + line
        carry = ''
        for match in find_articles(cline):
            art, word, eol_art = match.groups()
            if art is not None:
                assert word is not None
                yield (Location(file, i), art, word)
            else:
                assert eol_art is not None
                carry = eol_art + ' '

__all__ = ['parse_file']

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2022 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
import errno
import io
import os
import sys
import tempfile
import unittest.mock

from tests.tools import (
    assert_equal,
    assert_is_instance,
    assert_not_equal,
    isolation,
    testcase,
)

@contextlib.contextmanager
def tmpcwd():
    with tempfile.TemporaryDirectory(prefix='anorack.tests.') as tmpdir:
        orig_cwd = os.getcwd()
        os.chdir(tmpdir)
        try:
            yield
        finally:
            os.chdir(orig_cwd)

def TextIO(s=None, *, name):
    fp = io.BytesIO(s)
    fp.name = name
    return io.TextIOWrapper(fp, encoding='UTF-8')

class CompletedProcess():
    def __init__(self, rc, stdout, stderr):
        self.rc = rc
        self.stdout = stdout
        self.stderr = stderr

def __run_main(argv, stdin):
    sys.argv = argv
    if stdin is not None:
        if isinstance(stdin, str):
            stdin = stdin.encode('UTF-8')
        sys.stdin = mock_stdin = TextIO(stdin, name=sys.__stdin__.name)
    else:
        mock_stdin = None
    sys.stdout = mock_stdout = TextIO(name=sys.__stdout__.name)
    sys.stderr = mock_stderr = TextIO(name=sys.__stderr__.name)
    import lib.cli  # pylint: disable=bad-option-value,import-outside-toplevel
    rc = 0
    try:
        lib.cli.main()
    except SystemExit as exc:
        rc = exc.code
    except OSError as exc:
        rc = exc
    yield rc
    for fp in (sys.stdout, sys.stderr):
        fp.flush()
        s = fp.buffer.getvalue()  # pylint: disable=no-member
        yield s.decode('UTF-8')
    del mock_stdin, mock_stdout, mock_stderr

def _run_main(argv, stdin):
    # abuse mock to save&restore sys.argv, sys.stdin, etc.:
    with unittest.mock.patch.multiple(sys, argv=None, stdin=None, stdout=None, stderr=None):
        return CompletedProcess(*__run_main(argv, stdin))

run_main = isolation(_run_main)

def t(*, stdin=None, files=None, stdout, stdout_ipa=None, stderr='', stderr_ipa=None):
    if stdout_ipa is None:
        stdout_ipa = stdout
    if stderr_ipa is None:
        stderr_ipa = stderr
    argv = ['anorack']
    if files is not None:
        for (name, content) in files:
            with open(name, 'wt', encoding='UTF-8') as file:
                file.write(content)
            argv += [name]
    actual = run_main(argv, stdin)
    if '-@' in stdout:
        stdout = stdout.replace('-@', '@')
        actual.stdout = actual.stdout.replace('-@', '@')
    assert_equal(stdout, actual.stdout)
    assert_equal(stderr, actual.stderr)
    assert_equal(actual.rc, 0)
    argv += ['--ipa']
    actual = run_main(argv, stdin)
    actual.stderr = actual.stderr.replace('t͡ʃ', 'tʃ')
    assert_equal(stdout_ipa, actual.stdout)
    assert_equal(stderr_ipa, actual.stderr)
    assert_equal(actual.rc, 0)

@testcase
def test_stdin():
    t(
        stdin=(
            'It could be carried by an African swallow!\n'
            'Oh, yeah, a African swallow maybe, but not an\n'
            'European swallow.\n'
        ),
        stdout=(
            "<stdin>:2: a African -> an African /'afrIk@n/\n"
            "<stdin>:3: an European -> a European /j,U@r-@p'i@n/\n"
        ),
        stdout_ipa=(
            "<stdin>:2: a African -> an African /ˈafɹɪkən/\n"
            "<stdin>:3: an European -> a European /jˌʊəɹəpˈiən/\n"
        ),
    )

@testcase
@tmpcwd()
def test_files():
    t(
        files=(
            ('holy', 'It could be carried by a African swallow!'),
            ('grail', 'Oh, yeah, an African swallow maybe, but not an European swallow.'),
        ),
        stdout=(
            "holy:1: a African -> an African /'afrIk@n/\n"
            "grail:1: an European -> a European /j,U@r-@p'i@n/\n"
        ),
        stdout_ipa=(
            "holy:1: a African -> an African /ˈafɹɪkən/\n"
            "grail:1: an European -> a European /jˌʊəɹəpˈiən/\n"
        ),
    )

@testcase
def test_warning():
    def dummy_choose_art(phon):  # pylint: disable=unused-argument
        return NotImplemented
    with unittest.mock.patch('lib.cli.choose_art', dummy_choose_art):
        t(
            stdin='A scratch?!',
            stdout='',
            stderr="anorack: warning: can't determine correct article for 'scratch' /skr'atS/\n",
            stderr_ipa="anorack: warning: can't determine correct article for 'scratch' /skɹˈatʃ/\n",
        )

@testcase
def test_bad_io():
    argv = ['anorack', '/nonexistent', '-']
    actual = run_main(argv, 'a African')
    assert_equal('<stdin>:', actual.stdout[:8])
    err = os.strerror(errno.ENOENT)
    stderr = f'{argv[0]}: {argv[1]}: {err}\n'
    assert_equal(stderr, actual.stderr)
    assert_equal(actual.rc, 1)
    argv[1:1] = ['--traceback']
    actual = run_main(argv, 'a African')
    assert_equal('', actual.stdout)
    assert_equal('', actual.stderr)
    assert_is_instance(actual.rc, OSError)
    assert_equal(actual.rc.errno, errno.ENOENT)

@testcase
def test_changelog():
    argv = ['anorack', 'doc/changelog']
    actual = run_main(argv, None)
    assert_equal('', actual.stdout)
    assert_equal('', actual.stderr)
    assert_equal(actual.rc, 0)

@testcase
def test_version():
    argv = ['anorack', '--version']
    actual = run_main(argv, None)
    assert_not_equal('', actual.stdout)
    assert_equal('', actual.stderr)
    assert_equal(actual.rc, 0)

del testcase

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2021 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from tests.tools import (
    assert_equal,
    testcase,
)

import lib.espeak as M

def cmp(x, y):
    return (x > y) - (x < y)

@testcase
def test():
    versions = '''
    1
    1.04.71
    1.47.08
    1.48.1
    1.48.11
    1.48.15
    3
    3.14
    '''.split()
    for i1, v1 in enumerate(versions):
        for i2, v2 in enumerate(versions):
            assert_equal(M.vcmp(v1, v2), cmp(i1, i2))

del testcase

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2021 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest.mock

from tests.tools import (
    assert_equal,
    testcase,
)

import lib.io as M

def t(src, dst):
    class mock_stdout:
        encoding = src
    with unittest.mock.patch('sys.stdout', mock_stdout):
        encoding = M.get_encoding()
    assert_equal(encoding, dst)

@testcase
def test_ascii():
    t('ANSI_X3.4-1968', 'UTF-8')
    t('US-ASCII', 'UTF-8')
    t('ASCII', 'UTF-8')

@testcase
def test_8bit():
    t('ISO-8859-2', 'ISO-8859-2')

@testcase
def test_utf8():
    t('UTF-8', 'UTF-8')

del testcase

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2021 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io

from tests.tools import (
    assert_equal,
    assert_is,
    assert_is_instance,
    testcase,
)

import lib.parser as M

def t(s, exp):
    if isinstance(exp, tuple):
        exp = [exp]
    file = io.StringIO(s)
    result = list(M.parse_file(file))
    assert_equal(len(result), len(exp))
    for (loc, art, word), (xi, xart, xword) in zip(result, exp):
        assert_is_instance(loc, M.Location)
        assert_is(loc.file, file)
        assert_equal(loc.lineno, xi)
        assert_equal(xart, art)
        assert_equal(xword, word)

@testcase
def test_mid_line():
    t('Oh, yeah, an African swallow maybe,\nbut not a European swallow.\n', [
        (1, 'an', 'African'),
        (2, 'a', 'European'),
    ])

@testcase
def test_wrapped():
    t('I thought we were an\nautonomous collective.', (2, 'an', 'autonomous'))

@testcase
def test_quotes():
    t(
        "a 'scratch'\n"
        'a ‘scratch’\n'
        'a "scratch"\n'
        'a “scratch”\n',
        [(i, 'a', 'scratch') for i in range(1, 5)]
    )

@testcase
def test_underscore():
    t('a European_swallow', (1, 'a', 'European'))

@testcase
def test_numbers():
    t('an 8', (1, 'an', '8'))
    t('an 8bit', (1, 'an', '8'))
    t('an 8-bit', (1, 'an', '8'))
    t('an 8 bit', (1, 'an', '8'))
    t('a UTF16', (1, 'a', 'UTF'))
    t('a UTF-16', (1, 'a', 'UTF'))
    t('a UTF 16', (1, 'a', 'UTF'))

@testcase
def test_inner_apostrophe():
    t("Esra'a Al Shafei", [])
    t('Esra’a Al Shafei', [])

del testcase

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2016-2021 Jakub Wilk <jwilk@jwilk.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from tests.tools import (
    assert_equal,
    isolation,
    testcase,
    TestCase,
)

import lib.phonetics as M

def __test(word, xphon, xipa):
    M.init()
    phon = M.text_to_phonemes(word)
    assert_equal(xphon, phon)
    ipa = M.text_to_phonemes(word, ipa=True)
    ipa = ipa.replace('t͡ʃ', 'tʃ')
    assert_equal(xipa, ipa)
_test = isolation(__test)

class test_overrides(TestCase):
    pass

def _init_test_overrides():
    def add(word, xphon, xipa):
        @staticmethod
        def test():
            _test(word, xphon, xipa)
        setattr(test_overrides, 'test_' + word, test)
    add('EWMH', ",i:d,Vb@Lj,u:,Em'eItS", 'ˌiːdˌʌbəljˌuːˌɛmˈeɪtʃ')
    add('UCS', "j,u:s,i:;'Es", 'jˌuːsˌiːˈɛs')
    add('UDP', "j,u:d,i:p'i:", 'jˌuːdˌiːpˈiː')
    add('UPS', "j,u:p,i:;'Es", 'jˌuːpˌiːˈɛs')
    add('UTF', "j,u:t,i:;'Ef", 'jˌuːtˌiːˈɛf')
    add('UTS', "j,u:t,i:;'Es", 'jˌuːtˌiːˈɛs')
    add('UUID', "j,u:j,u:,aId'i:", 'jˌuːjˌuːˌaɪdˈiː')
    add('src', "s'o@s", 'sˈɔːs')
    add('unary', "j'un@ri", "jˈunəɹi")
    add('usr', "j,u:,Es'A@", "jˌuːˌɛsˈɑː")
_init_test_overrides()
del _init_test_overrides

del testcase

# vim:ts=4 sts=4 sw=4 et
